class DepGraph:
    """
    Based on a simple map, this class holds the name of each _model and
    contains a set with all of the dependent models (only needed when _model
    references are present)
    """

    def __init__(self):
        self.dep_map = {}

    def add_dependency(self, lib: str, dependents: "list[str]"):
        """Add a dependency to the graph

        Args:
            lib: name of the _model to add
            dependents: list of all dependents
        """
        if lib not in self.dep_map:
            self.dep_map.update({lib: set(dependents)})
        else:
            self.dep_map.update({lib: self.dep_map[lib].union(dependents)})

    def __contains__(self, obj):
        return obj in self.dep_map



from pysimlink import Model, print_all_params
import argparse
import numpy as np
from multiprocessing import Process
import os
import time


def runner(args):
    model = Model(args.model_name, args.model_path, force_rebuild=True)
    model.reset()

    for i in range(len(model)):
        model.step()
        print(f"here i: {i:5d} pid: {os.getpid():5d}")
        time.sleep(2)


def main(args):
    # procs = []
    # for i in range(10):
    #     this_proc = Process(target=runner, args=(args,))
    #     procs.append(this_proc)

    # list(map(lambda f: f.start(), procs))
    # list(map(lambda f: f.join(), procs))
    tic = time.time()
    model = Model(args.model_name, args.model_path, force_rebuild=True)
    toc = time.time()
    print("compiling took", toc - tic)
    model.reset()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("model_name")
    parser.add_argument("model_path")
    args = parser.parse_args()
    main(args)
    # tmp(args)
    # runner(args)



import os
import sys
import warnings

import numpy as np
from fasteners import InterProcessReaderWriterLock
import tempfile

from pysimlink.lib.model_paths import ModelPaths
from pysimlink.utils import annotation_utils as anno
from pysimlink.utils.model_utils import mt_rebuild_check, sanitize_model_name
from pysimlink.lib.model_types import DataType, ModelInfo
from pysimlink.lib.spinner import open_spinner
import pickle
import time
import importlib


class Model:
    """
    Instance of the simulink mode. This class compiles and imports
    the model once built. You can have multiple instances of the same
    model in one python runtime.
    """

    _model_paths: "anno.ModelPaths"
    _compiler: "anno.Compiler"

    def __init__(  # pylint: disable=R0913
        self,
        model_name: str,
        path_to_model: str,
        compile_type: str = "grt",
        suffix: str = "rtw",
        tmp_dir: "anno.Optional[str]" = None,
        force_rebuild: bool = False,
    ):
        """
        Args:
            model_name (str): name of the root simulink model
            path_to_model (str): path to the directory containing code generated from the Simulink Coder *or* the packaged zip file
            compile_type (str): Makefile template used to generate code from Simulink Coder. Only GRT is supported.
            suffix (str): Simulink Coder folders are almost always suffixed with rtw (real time workshop).
            tmp_dir (Optional[str]): Path to the directory that will be used to build the model. Defaults to :file:`__pycache__/{model_name}`
            force_rebuild (bool): force pysimlink to recompile the model from the source located at :code:`path_to_model`. Removes all build artifacts.

        Attributes:
            orientations: enumeration describing matrix orientations (row major, column major, etc.). This enumeration is
                likely the same among all models, but could change across MATLAB versions.
        """

        self._model_paths = ModelPaths(path_to_model, model_name, compile_type, suffix, tmp_dir)
        self._compiler = self._model_paths.compiler_factory()
        self._lock = InterProcessReaderWriterLock(
            os.path.join(tempfile.gettempdir(), model_name + ".lock")
        )
        with self._lock.write_lock():
            # Check need to compile
            if (
                mt_rebuild_check(self._model_paths, force_rebuild)
                or self._compiler.needs_to_compile()
            ):
                # Need to compile
                with open_spinner("Compiling"):
                    self._compiler.compile()
                with open(os.path.join(self._model_paths.tmp_dir, "compile_info.pkl"), "wb") as f:
                    obj = {"pid": os.getpid(), "parent": os.getppid(), "time": time.time()}
                    pickle.dump(obj, f)

        self.path_dirs = []
        for dir, _, _ in os.walk(
            os.path.join(self._model_paths.tmp_dir, "build", "out", "library")
        ):
            sys.path.append(dir)
            self.path_dirs.append(dir)

        self.module = importlib.import_module(self._model_paths.module_name)
        model_class = getattr(
            self.module, sanitize_model_name(self._model_paths.root_model_name) + "_Model"
        )

        self._model = model_class(self._model_paths.root_model_name)

        self.orientations = getattr(
            self.module,
            sanitize_model_name(self._model_paths.root_model_name) + "_rtwCAPI_Orientation",
        )

    def __del__(self):
        if sys.path is not None and hasattr(self, "path_dirs"):
            for dir in self.path_dirs:
                sys.path.remove(dir)
        # if hasattr(self, "module"):
        # del sys.modules[self.module]
        # sys.modules.remove(self.module)

    def __len__(self):
        """
        Get the total number of steps this model can run
        """
        return int(self.tFinal / self.step_size)

    def get_params(self) -> "list[anno.ModelInfo]":
        """
        Return an instance of all parameters, blocks, and signals in the _model

        See :func:`pysimlink.print_all_params` for iterating and printing the contents of this object

        Returns:
            list[:class:`pysimlink.types.ModelInfo`]: List of model info, one for each model (if reference models present). One ModelInfo if no reference models
        """
        return list(map(ModelInfo, self._model.get_params()))

    def reset(self):
        """
        Reset the simulink model. This clears all signal values and reinstantiates the model.
        """
        self._model.reset()

    def step(self, iterations: int = 1):
        """
        Step the simulink model

        Args:
            iterations: Number of timesteps to step internally.
                :code:`model.step(10)` is equivalent to calling :code:`for _ range(10): model.step(1)` functionally, but compiled.

        Raises:
            RuntimeError: If the model encounters an error (these will be raised from simulink). Most commonly, this
                will be `simulation complete`.

        """
        self._model.step(iterations)

    @property
    def tFinal(self) -> float:
        """
        Get the final timestep of the model.

        Returns:
            float: Final timestep of the model (seconds from zero).
        """
        return self._model.tFinal()

    @property
    def step_size(self) -> float:
        """
        Get the step size of the model

        Returns:
            float: step size of the fixed step solver.
        """
        return self._model.step_size()

    def set_tFinal(self, tFinal: float):
        """
        Change the final timestep of the model

        Args:
            tFinal: New final timestep of the model (seconds from zero).

        Raises:
            ValueError: if tFinal is <= 0
        """
        if tFinal <= 0:
            raise ValueError("new tFinal must be > 0")
        self._model.set_tFinal(tFinal)

    def get_signal(self, block_path, model_name=None, sig_name="") -> "np.ndarray":
        """
        Get the value of a signal

        Args:
            block_path: Path to the originating block
            model_name: Name of the model provided by :func:`pysimlink.print_all_params`. None if there are no model
                references (using :code:`None` will retrieve from the root model).
            sig_name: Name of the signal

        Returns:
            Value of the signal at the current timestep
        """
        model_name = self._model_paths.root_model_name if model_name is None else model_name
        return self._model.get_signal(model_name, block_path, sig_name)

    def get_block_param(self, block_path, param, model_name=None) -> "np.ndarray":
        """
        Get the value of a block parameter

        Args:
            block_path: Path the block within the model
            param: Name of the parameter to retrieve
            model_name: Name of the model provided by :func:`pysimlink.print_all_params`. None if there are no model references.

        Returns:
            np.ndarray with the value of the parameter
        """
        model_name = self._model_paths.root_model_name if model_name is None else model_name
        return self._model.get_block_param(model_name, block_path, param)

    def get_model_param(self, param, model_name=None) -> "np.ndarray":
        """
        Get the value of a model parameter

        Args:
            param: Name of the parameter to retrieve
            model_name: Name of the model provided by :func:`pysimlink.print_all_params`. None if there are no model references.

        Returns:
            np.ndarray with the value of the parameter
        """
        model_name = self._model_paths.root_model_name if model_name is None else model_name
        return self._model.get_model_param(model_name, param)

    def get_models(self) -> "list[str]":
        """
        Gets a list of all reference models (and the root model) in this model.

        Returns:
            list of paths, one for each model
        """
        return self._model.get_models()

    def set_block_param(
        self,
        block: str,
        param: str,
        value: "anno.ndarray",
        model_name: "anno.Union[str,None]" = None,
    ):
        """
        Set the parameter of a block within the model.

        Args:
            block: Path to the block within the model
            param: Name of the parameter to change
            value: new value of the parameter
            model_name: Name of the model provided by :func:`pysimlink.print_all_params`. None if there are no model references.
        Raises:
            RuntimeError: If the value array is not the correct shape or orientation as the parameter to change
        """
        model_name = self._model_paths.root_model_name if model_name is None else model_name
        info = self._model.block_param_info(model_name, block, param)
        dtype = DataType(info)
        if str(value.dtype) != dtype.pythonType:
            warnings.warn(
                f"Datatype of value does not match parameter. Expected {dtype.pythonType} got {value.dtype}",
                RuntimeWarning,
                stacklevel=2,
            )
        if dtype.orientation in [self.orientations.col_major_nd, self.orientations.col_major]:
            value = np.asfortranarray(value, dtype=dtype.pythonType)
        elif dtype.orientation in [self.orientations.row_major_nd, self.orientations.row_major]:
            value = np.ascontiguousarray(value, dtype=dtype.pythonType)
        elif str(value.dtype) != dtype.pythonType:
            value = value.astype(dtype.pythonType)

        self._model.set_block_param(model_name, block, param, value)

    def set_model_param(
        self, param: str, value: "anno.ndarray", model_name: "anno.Union[str,None]" = None
    ):
        """
        Set a model parameter.

        Args:
            param: Name of the parameter to change
            value: new value of the parameter
            model_name: Name of the model provided by :func:`pysimlink.print_all_params`. None if there are no model references.
        Raises:
            RuntimeError: If the value array is not the correct shape or orientation as the parameter to change
        """
        model_name = self._model_paths.root_model_name if model_name is None else model_name
        info = self._model.model_param_info(model_name, param)
        dtype = DataType(info)
        if str(value.dtype) != dtype.pythonType:
            warnings.warn(
                f"Datatype of value does not match parameter. Expected {dtype.pythonType} got {value.dtype}",
                RuntimeWarning,
                stacklevel=2,
            )
        if dtype.orientation in [self.orientations.col_major_nd, self.orientations.col_major]:
            value = np.asfortranarray(value, dtype=dtype.pythonType)
        elif dtype.orientation in [self.orientations.row_major_nd, self.orientations.row_major]:
            value = np.ascontiguousarray(value, dtype=dtype.pythonType)
        elif str(value.dtype) != dtype.pythonType:
            value = value.astype(dtype.pythonType)

        self._model.set_model_param(model_name, param, value)



import os
import glob
import re

import pybind11


from pysimlink.utils.model_utils import sanitize_model_name


class CmakeTemplate:
    """
    Generates the CMakeLists.txt file that can be used to compile the model.

    The generated cmakelists.txt includes custom mixins and pybind bindings that
    allow a model to be imported in python.
    """

    def __init__(self, model_name):
        self.model_name = model_name
        self.libs = []
        self.replacers = [(re.compile(r"(?<!\\) "), r"\ "), (re.compile(r"\\(?! )"), r"/")]

        self.sanitized_name = sanitize_model_name(model_name)

    def replacer(self, string):
        for search, rep in self.replacers:
            indices = list(re.finditer(search, string))
            for idx in reversed(indices):
                string = string[: idx.start(0)] + rep + string[idx.end(0) :]
        return string

    def header(self):
        """
        Create the header of the CMakeLists.txt

        Returns:
            str: header section for cmake
        """
        return f"""cmake_minimum_required(VERSION 3.4)
project({self.model_name})
set(CMAKE_CXX_STANDARD 11)
set(CMAKE_CXX_STANDARD_REQUIRED ON)
find_package(pybind11 PATHS {self.replacer(pybind11.get_cmake_dir())})"""

    def set_includes(self, includes: "list[str]"):
        """
        Setup the includes directories for the model, custom mixins, and simulink files

        Args:
            includes: list of all include paths.

        Returns:
            str: include_directories directive
        """
        ## Add the include path for model runner code

        includes = [self.replacer(os.path.abspath(include)) for include in includes]
        include_dirs = "\n    ".join(includes)
        return f"""
include_directories(
    {include_dirs}
    ${{pybind11_INCLUDE_DIRS}}
    #${{PYTHON_INCLUDE_DIRS}}
)
"""

    def add_library(self, lib_name: str, sources: "list[str]"):
        """
        Add a single library to be compiled

        Args:
            lib_name: Name of the library
            sources: list of all source files to compile into this library

        Returns:
            str: add_library cmake directive
        """
        self.libs.append(lib_name)
        sources = [self.replacer(os.path.abspath(source)) for source in sources]
        source_paths = "\n        ".join(sources)
        return f"""
add_library(
    {lib_name}
        {source_paths}
)        
"""

    def set_lib_props(self):
        """
        Set all models to be compiled with c 90

        This is usually fine, even if the model was generated to use c89. Not the other way around though.

        Returns:
            str: set_target_properties cmake directive
        """
        libs = "\n    ".join(self.libs)
        return f"""
set_target_properties(
    {libs}
        PROPERTIES C_STANDARD 90
)        
"""

    def add_custom_libs(self, sources: "list[str]"):
        """
        Add the mixins library (created with pybind11)

        Args:
            sources: list of all source files part of the library

        Returns:
            str: pybind11_add_module cmake directive
        """
        sources = glob.glob(sources + "/*.cpp")

        sources = [self.replacer(os.path.abspath(source)) for source in sources]
        source_paths = "\n        ".join(sources)
        return f"""
pybind11_add_module(
    {self.sanitized_name}_interface_c
        {source_paths}
)        

set_target_properties(
    {self.sanitized_name}_interface_c PROPERTIES
        LIBRARY_OUTPUT_DIRECTORY ${{PROJECT_BINARY_DIR}}/out/library
)
"""

    def add_link_libs(self, dep_map: "dict[str, list[str]]"):
        """
        Link libraries according to the dependency map

        Args:
            dep_map: dictionary from the dependency_graph class

        Returns:
            str: all required target_link_libraries cmake directives
        """
        ret = ""
        for dep, deps in dep_map.items():
            if len(deps) == 0:
                continue
            deps_exp = "\n        ".join(deps)
            ret += f"""
target_link_libraries(
    {dep}
        {deps_exp}
)
"""
        return ret

    def add_private_link(self, root_model: "str"):
        """
        Link the custom mixins to the root model

        Args:
            root_model: name of the root model

        Returns:
            str: target_link_libraries directive, but as private link
        """
        return f"""
target_link_libraries(
    {self.sanitized_name}_interface_c
        PRIVATE {root_model}
)        
"""

    def footer(self):
        """
        Set position independent code flag for all targets

        Returns:
            str: set directives for all targets
        """
        return """
set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -fPIC")
set(CMAKE_C_FLAGS "${CMAKE_C_FLAGS} -fPIC")
"""

    def add_compile_defs(self, defines: "list[str]"):
        """
        Set compile definitions from defines.txt or inference

        Args:
            defines: list of defines added to cmake

        Returns:
            str: add_compile_definitions cmake directive
        """
        defines = "\n    ".join(defines)
        return f"""
add_compile_definitions(
    {defines}
)        
"""



from dataclasses import dataclass
from pysimlink.utils import annotation_utils as anno


# These classes are provided for reference only.
@dataclass
class DataType:
    """
    Contains data type information for a single parameter.

    Attributes:
        cDataType (str): The c equivalent name of this datatype
        pythonType (str): The python name of this datatype
        dims (list[int]): List of dimension sizes. (the shape of the array)
        orientation (int): Enumeration indicating if this is a scalar, vector, column major or row major array.
            Since this enum is evaluated at compile time, it is imported from the compiled model. To check this value
            dynamically, use :attr:`pysimlink.Model.orientations`.
    """

    cDataType: str
    pythonType: str
    dims: "list[int]"
    orientation: int

    def __init__(self, obj: "anno.c_model_datatype"):
        self.cDataType = obj.cDataType
        self.pythonType = obj.pythonType
        self.dims = obj.dims
        self.orientation = obj.orientation

    def __repr__(self):
        return f"{self.pythonType} ({self.cDataType}) dims: {self.dims} order: {self.orientation}"


@dataclass
class ModelParam:
    """
    Represents a single model parameter.

    Attributes:
        model_param (str): Name of the model parameter (these are usually stored withing the model workspace in simulink)
        data_type (:class:`DataType`): data type object describing this parameter
    """

    model_param: str
    data_type: DataType

    def __init__(self, obj: "anno.c_model_param"):
        self.model_param = obj.model_param
        self.data_type = DataType(obj.data_type)


@dataclass
class BlockParam:
    """
    Represents a single parameter in a block.

    If a block has *n* parameters, there will be *n* BlockParam instances. Each with the same :code:`block_name`.

    Attributes:
        block_name (str): The name (usually including the path) to this block.
        block_param (str): The name of the parameter within the block
        data_type (:class:`DataType`): data type object describing this parameter
    """

    block_name: str
    block_param: str
    data_type: DataType

    def __init__(self, obj: "anno.c_model_block_param"):
        self.block_name = obj.block_name
        self.block_param = obj.block_param
        self.data_type = DataType(obj.data_type)


@dataclass
class Signal:
    """
    Represents a single signal in the model.

    Attributes:
        block_name: name and path to the block the signal originates from
        signal_name: name of the signal (empty if not named)
        data_type (:class:`DataType`): Data type information of this signal
    """

    block_name: str
    signal_name: str
    data_type: DataType

    def __init__(self, obj: "anno.c_model_signal"):
        self.block_name = obj.block_name
        self.signal_name = obj.signal_name
        self.data_type = DataType(obj.data_type)


@dataclass
class ModelInfo:
    """
    Object returned from :func:`pysimlink.Model.get_params`.

    .. note:: This object only describes a single model. If a Simulink model contains model references,
        :func:`pysimlink.Model.get_params` will return a list of this object, each describing a reference model.

    Attributes:
        model_name (str): Name of the model this object describes
        model_params (list[:class:`ModelParam`]): List of all model parameters
        block_params (list[:class:`BlockParam`]): List of all block parameters
        signals (list[:class:`Signal`]): List of all signals
    """

    model_name: str
    model_params: "list[ModelParam]"
    block_params: "list[BlockParam]"
    signals: "list[Signal]"

    def __init__(self, obj: "anno.c_model_info"):
        self.model_name = obj.model_name
        self.model_params = list(map(ModelParam, obj.model_params))
        self.block_params = list(map(BlockParam, obj.block_params))
        self.signals = list(map(Signal, obj.signals))



import typing

if typing.TYPE_CHECKING:
    from pysimlink.lib.model_types import ModelInfo
    from pysimlink.lib.dependency_graph import DepGraph
    from pysimlink.lib.model_paths import ModelPaths
    from pysimlink.lib.compilers.compiler import Compiler
    from pysimlink.lib.model import Model
    from typing import Union
    from numpy import ndarray
    from typing import Optional

    c_model_info = typing.Any
    c_model_param = typing.Any
    c_model_datatype = typing.Any
    c_model_signal = typing.Any
    c_model_block_param = typing.Any



class GenerationError(Exception):
    """
    Exception raised when generating build files fails.

    Attributes:
        dump (str): stdout and stderr output of the **generation** process
        cmake (str): contents of the :file:`CMakeLists.txt` file
    """

    def __init__(self, *args):
        self.dump = args[0]
        self.cmake = args[1]

    def __str__(self):
        return (
            "Generating the CMakeLists for this model failed. This could be a c/c++/cmake setup issue, bad paths, or a bug! "
            f"Output from CMake generation is in {self.dump}"
        )


class BuildError(Exception):
    """
    Exception raised when generating build files succeeded but compiling the model fails.

    Attributes:
        dump (str): stdout and stderr output of the **build** process
        cmake (str): contents of the :file:`CMakeLists.txt` file
    """

    def __init__(self, *args):
        self.dump = args[0]
        self.cmake = args[1]

    def __str(self):
        return (
            "Building the model failed. This could be a c/c++/cmake setup issue, bad paths, or a bug! "
            f"Output from the build process is in {self.dump}"
        )



import os
import re
import glob

from pysimlink.lib.dependency_graph import DepGraph
from pysimlink.lib.cmake_gen import CmakeTemplate
from pysimlink.lib.compilers.compiler import Compiler
from pysimlink.utils import annotation_utils as anno


class ModelRefCompiler(Compiler):
    """
    Compiler for a model that do use model references
    """

    def __init__(self, model_paths: "anno.ModelPaths"):
        super().__init__(model_paths)
        self.models = None

    def compile(self):
        self.clean()
        self._get_simulink_deps()
        self._build_deps_tree()
        self._gen_custom_srcs()
        self._gen_cmake()
        self._build()

    def _build_deps_tree(self):
        """
        Get the dependencies for this _model and all child models recursively,
        starting at the root _model.

        """
        self.models = DepGraph()
        self.update_recurse(self.model_paths.root_model_name, self.models, is_root=True)

    def update_recurse(self, model_name: str, models: "anno.DepGraph", is_root=False):
        """
        Recursively gathers dependencies for a _model and adds them to the
        dependency graph.
        Argument:
            model_name: Name of the _model. This is used to locate the _model in
                the slprj/{compile_type} folder
            models: An instance of the dependency graph
            is_root: Since the root _model is in a different place, we handle the
                path differently. This is set to true only when processing the
                root _model
        Returns:
            None: always
        """
        if model_name in models:
            return

        model_path = (
            self.model_paths.root_model_path
            if is_root
            else os.path.join(self.model_paths.slprj_dir, model_name)
        )
        deps = self._get_deps(model_path, model_name)
        models.add_dependency(model_name, deps)
        for dep in deps:
            self.update_recurse(dep, models)

    def _get_deps(self, path, model_name):
        """Get all dependencies of a _model

        Args:
            path: Path to the _model (including it's name). Must contain
                {model_name}.h
            model_name: Name of the _model. path must contain {model_name}.h

        Returns:
            deps: list of dependencies for this _model.
        """
        deps = set()
        with open(os.path.join(path, model_name + ".h"), encoding="utf-8") as f:
            regex = re.compile('^#include "(.*?)"')
            end = re.compile("^typedef")
            for line in f.readlines():
                inc_test = re.match(regex, line)
                if inc_test:
                    dep = inc_test.groups()[0]
                    ## Could probably replace this with .split('.')[0] but can _model names have a '.'?
                    suffix_idx = dep.find(".h")
                    deps.add(dep[:suffix_idx])
                    continue
                if re.match(end, line):
                    break
        this_deps = deps.difference(self.simulink_deps)
        to_remove = [dep for dep in this_deps if dep.startswith(model_name)]
        return this_deps.difference(to_remove)

    def _get_simulink_deps(self):
        super()._get_simulink_deps()
        _shared_utils = glob.glob(
            os.path.join(self.model_paths.slprj_dir, "_sharedutils") + "/**/*.c",
            recursive=True,
        )
        self.simulink_deps_src += _shared_utils

        _shared_utils = glob.glob(
            os.path.join(self.model_paths.slprj_dir, "_sharedutils") + "/**/*.h",
            recursive=True,
        )
        self.simulink_deps = self.simulink_deps.union(
            [os.path.basename(f).split(".")[0] for f in _shared_utils]
        )

    def _gen_cmake(self):
        includes = [self.custom_includes]
        for dir_name in os.walk(self.model_paths.root_dir, followlinks=False):
            for file in dir_name[2]:
                if ".h" in file:
                    includes.append(dir_name[0])
                    break

        maker = CmakeTemplate(
            self.model_paths.root_model_name.replace(" ", "_").replace("-", "_").lower()
        )
        cmake_text = maker.header()
        cmake_text += maker.set_includes(includes)

        for lib in self.models.dep_map:
            if lib == self.model_paths.root_model_name:
                files = glob.glob(self.model_paths.root_model_path + "/*.c")
            else:
                files = glob.glob(os.path.join(self.model_paths.slprj_dir, lib) + "/*.c")
            cmake_text += maker.add_library(lib, files)

        cmake_text += maker.add_library("shared_utils", self.simulink_deps_src)
        ## the custom code depends on the root _model.
        self.models.add_dependency(self.model_paths.root_model_name, ["shared_utils"])

        cmake_text += maker.add_custom_libs(self.custom_sources)
        cmake_text += maker.add_private_link(self.model_paths.root_model_name)
        cmake_text += maker.set_lib_props()
        cmake_text += maker.add_link_libs(self.models.dep_map)
        cmake_text += maker.add_compile_defs(self.defines)
        cmake_text += maker.footer()

        with open(
            os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(cmake_text)

    @property
    def _module_name(self):
        return
        return super()._module_name



import glob
import os
import re
import sys
from typing import Union
import zipfile

from pysimlink.utils import annotation_utils as anno
from pysimlink.utils.model_utils import get_other_in_dir, sanitize_model_name


class ModelPaths:
    """
    Holds information about the paths to the model being built.
    """

    root_dir: str  ## Directory containing all model components.
    simulink_native: str  ## path to the directory containing stuff
    #  generated by simulink for every model
    root_model_path: str  ## Path to the root model
    root_model_name: str  ## Name of the root model
    compile_type: str  ## grt, ert, etc...
    suffix: str  ## Suffix added to the model name. Usually 'rtw'
    has_references: bool  ## If this model contains references
    models_dir: str  ## directory containing all simulink code related to the models
    slprj_dir: Union[str, None]  ## Directory will all child models (contains compile_type)
    tmp_dir: str  ## Directory where all compiled models will be built

    def __init__(
        self,
        root_dir: str,
        model_name: str,
        compile_type: str = "grt",
        suffix: str = "rtw",
        tmp_dir: "Union[str, None]" = None,
    ):
        """
        Args:
            root_dir: Directory created during codegen. Should have two directories in it.
            model_name: Name of the root model.
            compile_type: grt, ert, etc...
            suffix: the suffix added to the model name directory. usually 'rtw'
            tmp_dir: Where to store the build files. Defaults to __pycache__
        """
        self.compile_type = compile_type
        if self.compile_type != "grt":
            raise ValueError(
                "Unsupported compile target. grt is the only supported simulink "
                "code generation target.\nChange your code generation settings "
                "to use the grt.tlc target and try again. (compile_type should "
                f"be `grt` not {self.compile_type})"
            )
        self.suffix = suffix
        zip_test = os.path.splitext(root_dir)
        if zip_test[-1] == ".zip":
            with zipfile.ZipFile(root_dir, "r") as f:
                if tmp_dir is None:
                    ext_dir = os.path.join(
                        os.path.dirname(sys.argv[0]),
                        "__pycache__",
                        "extract",
                        os.path.basename(zip_test[0]),
                    )
                f.extractall(ext_dir)
            self.root_dir = ext_dir
        else:
            self.root_dir = root_dir

        walk = os.walk(self.root_dir, followlinks=False)
        for (cur_path, folders, _) in walk:
            if "simulink" in folders:
                self.simulink_native = cur_path
                break
        else:
            raise RuntimeError(f"{self.root_dir} is not a valid simulink model.")

        models_dir = get_other_in_dir(self.root_dir, os.path.basename(self.simulink_native))
        self.models_dir = os.path.join(self.root_dir, models_dir)

        self.has_references = os.path.exists(os.path.join(self.models_dir, "slprj"))

        self.root_model_path = os.path.join(
            self.models_dir, model_name + "_" + compile_type + "_" + suffix
        )
        if not os.path.exists(self.root_model_path):
            try:
                model_name = model_name.split("_" + compile_type + "_" + suffix)[0]
            except:  # pylint: disable=W0702
                pass
            self.root_model_path = os.path.join(
                self.models_dir, model_name + "_" + compile_type + "_" + suffix
            )
            if not os.path.exists(self.root_model_path):
                raise RuntimeError(
                    f"Cannot find folder with name '{model_name}' in '{self.models_dir}'"
                )
        self.root_model_name = model_name
        if self.has_references:
            self.slprj_dir = os.path.join(self.models_dir, "slprj", compile_type)

        if tmp_dir is None:
            self.tmp_dir = os.path.join(
                os.path.dirname(sys.argv[0]),
                "__pycache__",
                "pysimlink",
                self.root_model_name,
            )
        else:
            self.tmp_dir = os.path.join(tmp_dir, model_name)
        os.makedirs(self.tmp_dir, exist_ok=True)
        self.verify_capi()

    def verify_capi(self):
        """
        Make sure that this model was generated with the c api. This doesn't use
        the function in the capi, but we need the model mapping interface (mmi).
        """
        files = glob.glob(self.root_model_path + "/*.c", recursive=False)
        files = map(os.path.basename, files)
        assert self.root_model_name + "_capi.c" in files, (
            "Model not generated with capi. Enable the following options in the Code Generation model settings: \n"
            "\tGenerate C API for: signals, parameters, states, root-level I/O"
        )

        ## also check that this is not a multitasked model
        with open(
            os.path.join(self.root_model_path, self.root_model_name + ".h"), encoding="utf-8"
        ) as f:
            lines = f.readlines()

        regex = re.compile(
            f"extern void {self.root_model_name}_step\(void\);"  # pylint: disable=W1401
        )
        for line in lines:
            if re.search(regex, line):
                break
        else:
            raise RuntimeError(
                "Model is setup with multitasking. Disable the following options in the Solver settings and recompile: \n"
                "\t- 'Treat each discrete rate as a separate task'\n\t- 'Allow tasks to execute concurrently on target'"
            )

    def compiler_factory(self) -> "anno.Compiler":
        """
        Return the correct compiler. This could be simplified later -or- more
        compilers could be added if we want to use something other than cmake.
        """
        if self.has_references:  # pylint: disable=R1705
            from pysimlink.lib.compilers.model_ref_compiler import (  # pylint: disable=C0415
                ModelRefCompiler,
            )

            return ModelRefCompiler(self)
        else:
            from pysimlink.lib.compilers.one_shot_compiler import (  # pylint: disable=C0415
                NoRefCompiler,
            )

            return NoRefCompiler(self)

    @property
    def module_name(self):
        return sanitize_model_name(self.root_model_name) + "_interface_c"



from abc import abstractmethod
import glob
import os
import shutil
from datetime import datetime
from subprocess import Popen, PIPE

import cmake

from pysimlink.utils import annotation_utils as anno
from pysimlink.utils.model_utils import infer_defines, sanitize_model_name
from pysimlink.lib.exceptions import GenerationError, BuildError


class Compiler:
    """
    Base class to discover sources and compile a model
    """

    simulink_deps: "set[str]"  ## All header files created in the simulink common directory
    simulink_deps_path: "list[str]"  ## Path to the header files created in the simulink common dir
    simulink_deps_src: "list[str]"  ## Path to all source files created in the simulink common dir
    model_paths: "anno.ModelPaths"  ## Instance of the ModelPaths containing information about the directory structure
    defines: "list[str]"  ## All defines that should be set during _model compilation
    custom_includes: str  ## Include files directory defined by this python module
    custom_sources: str  ## Source files directory defined by this python module

    def __init__(self, model_paths: "anno.ModelPaths"):
        self.model_paths = model_paths

    def clean(self):
        """
        Remove all files from the temporary directory
        """
        shutil.rmtree(self.model_paths.tmp_dir, ignore_errors=True)

    def compile(self):
        """
        Builds the cmake file, calls cmake, and builds the extension.
        """
        raise NotImplementedError

    def needs_to_compile(self) -> bool:
        """
        check if the model extension exists.

        Returns:
            bool: True if the model needs to be compiled, False otherwise
        """
        if os.name == "nt":
            lib = glob.glob(
                os.path.join(
                    self.model_paths.tmp_dir,
                    "build",
                    "out",
                    "library",
                    "Debug",
                    self.model_paths.module_name + ".*",
                )
            )
        else:
            lib = glob.glob(
                os.path.join(
                    self.model_paths.tmp_dir,
                    "build",
                    "out",
                    "library",
                    self.model_paths.module_name + ".*",
                )
            )
        return len(lib) == 0

    def _get_simulink_deps(self):
        """
        Generates a list of all simulink dependencies and their paths
        """
        files = glob.glob(self.model_paths.simulink_native + "/**/*.h", recursive=True)

        self.simulink_deps = {os.path.basename(f).split(".")[0] for f in files}
        self.simulink_deps_path = files

        simulink_deps = glob.glob(self.model_paths.simulink_native + "/**/*.c", recursive=True)
        rt_main = None
        for file in simulink_deps:
            if os.path.basename(file) in ["rt_main.c", "classic_main.c"]:
                rt_main = file
                break
        if rt_main is not None:
            simulink_deps.remove(rt_main)

        self.simulink_deps_src = simulink_deps

    def _gen_custom_srcs(self):
        """
        Moves all custom mixin source files to the temporary directory and makes appropriate replacements
        in the source files
        """
        shutil.rmtree(
            os.path.join(self.model_paths.tmp_dir, "c_files"),
            ignore_errors=True,
        )
        shutil.copytree(
            os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "c_files")),
            os.path.join(self.model_paths.tmp_dir, "c_files"),
        )
        self.custom_includes = os.path.join(self.model_paths.tmp_dir, "c_files", "include")
        self.custom_sources = os.path.join(self.model_paths.tmp_dir, "c_files", "src")

        replacements = {
            "<<ROOT_MODEL>>": self.model_paths.root_model_name + ".h",
            "<<ROOT_MODEL_PRIVATE>>": self.model_paths.root_model_name + "_private.h",
            "<<MODEL_INTERFACE_C>>": self.model_paths.module_name,
            "<<ROOT_MODEL_NAME>>": sanitize_model_name(self.model_paths.root_model_name),
        }
        self._replace_macros(os.path.join(self.custom_includes, "model_utils.hpp"), replacements)
        self._replace_macros(
            os.path.join(self.custom_includes, "model_interface.hpp"), replacements
        )
        self._replace_macros(os.path.join(self.custom_sources, "bindings.cpp"), replacements)

        defines = os.path.join(self.model_paths.root_model_path, "defines.txt")
        if os.path.exists(defines):
            with open(defines, "r", encoding="utf-8") as f:
                self.defines = [line.strip() for line in f.readlines()]
        else:
            self.defines = infer_defines(self.model_paths)

    def _build(self):
        """
        Cals cmake to configure and build the extension. Writes errors to the current working directory
        in a log file.
        """
        build_dir = os.path.join(self.model_paths.tmp_dir, "build")

        with Popen(
            [
                os.path.join(cmake.CMAKE_BIN_DIR, "cmake"),
                "-S",
                self.model_paths.tmp_dir,
                "-DCMAKE_BUILD_TYPE=Release",
                "-B",
                build_dir,
            ],
            stdout=PIPE,
            stderr=PIPE,
        ) as p:
            (output1, err1) = p.communicate()
            build = p.wait()

        if build != 0:
            now = datetime.now()
            err_file = os.path.join(
                os.getcwd(),
                now.strftime("%Y-%m-%d_%H-%M-%S_PySimlink_Generation_Error.log"),
            )
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(output1.decode() if output1 else "")
                f.write(err1.decode() if err1 else "")
            raise GenerationError(
                err_file, os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt")
            )

        with Popen(
            [os.path.join(cmake.CMAKE_BIN_DIR, "cmake"), "--build", build_dir],
            stdout=PIPE,
            stderr=PIPE,
        ) as p:
            (output2, err2) = p.communicate()
            make = p.wait()

        if make != 0:
            now = datetime.now()
            err_file = os.path.join(
                os.getcwd(),
                now.strftime("%Y-%m-%d_%H-%M-%S_PySimlink_Build_Error.log"),
            )
            with open(err_file, "w", encoding="utf-8") as f:
                f.write(output2.decode() if output2 else "")
                f.write(err2.decode() if err2 else "")

            raise BuildError(err_file, os.path.join(self.model_paths.tmp_dir, "CMakeLists.txt"))

    @staticmethod
    def _replace_macros(path: str, replacements: "dict[str, str]"):
        """
        Replaces strings in a file

        Args:
            path: path to file to replace strings in
            replacements: dictionary of replacements

        """
        with open(path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            for key, val in replacements.items():
                lines[i] = lines[i].replace(str(key), str(val))

        with open(path, "w", encoding="utf-8") as f:
            f.writelines(lines)
